#lang sicp

(#%require "leval.rkt")
(#%require "test-eval.rkt")

(define (program delay)
  `((define count 0)
    (define (id ,(qualify-arg 'x delay))
      (set! count (+ count 1)) x)
    (define (square ,(qualify-arg 'x delay))
      (* x x))
    (square (id 10))
    count))

(define (input-sequence val count)
  (cons 'list
        (let loop ((n count))
          (if (= n 0) '()
              (cons val (loop (- n 1)))))))

(define (map-program length delay)
  `((define (f ,(qualify-arg 'x delay))
      (+ x (sqrt x)))
    (define input
      ,(input-sequence 5 length))
    (define (map ,(qualify-arg 'f delay)
                 ,(qualify-arg 'x delay))
      (if (null? x) '()
          (cons (f (car x))
                (map f (cdr x)))))
    (map f input)))


(#%require (only racket/base module+))
(#%require rackunit)  

(module+ test
  ;; With the normal lazy evaluator, square evaluates id twice
  (check-equal? (lazy-eval-program (program 'lazy)) 2)
  ;; With the memoizing evaluator, the first call to id is memoized,
  ;; so another call is not needed.
  (check-equal? (lazy-eval-program (program 'lazy-memo)) 1))

(module+ timing

  ;; Map needs to re-evaluate the procedure f for each element of the
  ;; input for lazy evaluation.  If memoized, this is only evaluated once.
  (define lazy-map (time-proc 100 (lambda () (lazy-eval-program
                                              (map-program 100 'lazy)))))
  (define lazy-memo-map (time-proc 100 (lambda () (lazy-eval-program
                                                   (map-program 100 'lazy-memo)))))
  (display `("Map: lazy " ,lazy-map " lazy-memo " ,lazy-memo-map " speed-up "
                          ,(/ lazy-map lazy-memo-map)))

  ;; Gives ~16x speedup.
)

;; In general, performance for the memoizing evaluator should be
;; much better when multiple references are made to the result of
;; evaluating a procedure.
