#lang sicp

(#%require "environment.rkt"
           "common.rkt"
           "syntax.rkt")

(#%provide eval-if
           eval-sequence
           eval-assignment
           eval-definition
           eval-unbind
           eval-and
           eval-or)

;; Evaulation of special forms

(define (eval-if exp env eval-proc)
  (if (true? (eval-proc (if-predicate exp) env))
      (eval-proc (if-consequent exp) env)
      (eval-proc (if-alternative exp) env)))

(define (eval-sequence exps env eval-proc)
  (cond ((last-exp? exps) (eval-proc (first-exp exps) env))
        (else (eval-proc (first-exp exps) env)
              (eval-sequence (rest-exps exps) env eval-proc))))

(define (eval-assignment exp env eval-proc)
  (set-variable-value! (assignment-variable exp)
                       (eval-proc (assignment-value exp) env)
                       env)
  'ok)

(define (eval-definition exp env eval-proc)
  (define-variable! (definition-variable exp)
    (eval-proc (definition-value exp) env)
    env)
  'ok)

(define (eval-unbind exp env eval-proc)
  (make-unbound-variable! (unbound-variable exp) env))

(define (eval-and conjuncts env eval-proc)
  (cond ((null? conjuncts)
         'true)
        ((last-conjunct? conjuncts)
         (eval-proc (first-conjunct conjuncts) env))
        ((false? (eval-proc (first-conjunct conjuncts) env))
         'false)
        (else
         (eval-and (make-and (rest-conjuncts conjuncts))
                   env eval-proc))))

(define (eval-or disjuncts env eval-proc)
  (if (null? disjuncts)
      'false
      (let ((val (eval-proc (first-disjunct disjuncts) env)))
        (if val
            val
            (eval-or (make-or (rest-disjuncts disjuncts))
                     env eval-proc)))))



