#lang sicp

;; Exercise 4.24: Define a procedure to run a program through an
;; interpreter and time the run.

(#%require "dd-mceval.rkt"
           "analyzing-mceval.rkt"
           "leval.rkt")

(#%provide qualify-arg
           time-proc)

(#%require (only racket current-process-milliseconds))
(#%require (only racket/base module+))


;; Call proc on args n times and output the average time per run
(define time-proc
  (lambda (iterations proc)
    (let ((start-time (current-process-milliseconds)))
      (let loop ((n iterations))
        (if (> n 0)
            (begin
              (proc)
              (loop (- n 1)))))
      (exact->inexact
       (/ (- (current-process-milliseconds)
             start-time)
          iterations)))))

(define (fact-program x)
  `((define (fact n)
      (if (= n 1) 1
          (* n (fact (- n 1)))))
    (fact ,x)))


(define (qualify-arg arg delay)
  (cond ((eq? delay 'strict) arg)
        ((eq? delay 'lazy) `(,arg lazy))
        ((eq? delay 'lazy-memo) `(,arg lazy-memo))
        (else (error "Unknown delay" delay))))

(define (fib-program x delay)
  `((define (fib
             ,(qualify-arg 'n delay))
      (cond ((= n 0) 0)
            ((= n 1) 1)
            (else
             (+ (fib (- n 2))
                (fib (- n 1))))))
    (fib ,x)))

(define (lazy-test-program delay)
  `((define (try a
                 ,(qualify-arg 'b delay))
      (if (= a 0) 1 b))
    (try 0 (/ 1 0))))

(#%require rackunit)

(module+ test
  (check-equal? (dd-eval-program (fact-program 1)) 1)
  (check-equal? (analyzing-eval-program (fact-program 1)) 1)
  (check-equal? (lazy-eval-program (fact-program 1)) 1)
  (check-equal? (dd-eval-program (fact-program 5)) 120)
  (check-equal? (analyzing-eval-program (fact-program 5)) 120)
  (check-equal? (lazy-eval-program (fact-program 5)) 120)
  (check-equal? (dd-eval-program (fact-program 10))
                (analyzing-eval-program (fact-program 10)))
  (check-equal? (dd-eval-program (fib-program 5 'strict))
                (analyzing-eval-program (fib-program 5 'strict)))
  (check-equal? (dd-eval-program (fib-program 5 'strict))
                (lazy-eval-program (fib-program 5 'strict)))
  (check-equal? (lazy-eval-program (fib-program 5 'strict))
                (lazy-eval-program (fib-program 5 'lazy)))
  (check-equal? (lazy-eval-program (fib-program 5 'strict))
                (lazy-eval-program (fib-program 5 'lazy-memo)))
  (check-equal? (lazy-eval-program (lazy-test-program 'lazy)) 1)
  (check-equal? (lazy-eval-program (lazy-test-program 'lazy-memo)) 1))

(module+ timing
  (define dd-fib (time-proc 100 (lambda () (dd-eval-program
                                            (fib-program 10 'strict)))))
  (define analyzing-fib (time-proc 100 (lambda () (analyzing-eval-program
                                                   (fib-program 10 'strict)))))
  (define dd-fact (time-proc 100 (lambda () (dd-eval-program
                                             (fact-program 50)))))
  (define analyzing-fact (time-proc 100 (lambda () (analyzing-eval-program
                                                    (fact-program 50)))))

  (define lazy-fib (time-proc 100 (lambda () (lazy-eval-program
                                              (fib-program 10 'lazy)))))
  (define lazy-memo-fib (time-proc 100 (lambda () (lazy-eval-program
                                                   (fib-program 10 'lazy-memo)))))

  (display `("Fib: dd " ,dd-fib " analyzing " ,analyzing-fib " speed-up "
                        ,(/ dd-fib analyzing-fib)))
  (newline)
  (display `("Fact: dd " ,dd-fact " analyzing " ,analyzing-fact " speed-up "
                         ,(/ dd-fib analyzing-fact)))
  (newline)
  (display `("Fib: lazy " ,lazy-fib " memoizing " ,lazy-memo-fib " speed-up "
                          ,(/ lazy-fib lazy-memo-fib))))


;; Test results
;; test-eval.rkt> (time-proc 100 eval-program fib-program mce-eval mce-user-print)
;; 203/100
;; test-eval.rkt> (time-proc 100 eval-program fib-program analyzing-eval analyzing-user-print)
;; 18/25

;; test-eval.rkt> (time-proc 100 eval-program fact-program mce-eval mce-user-print)
;; 183/100
;; test-eval.rkt> (time-proc 100 eval-program fact-program analyzing-eval analyzing-user-print)
;; 89/100

;; This demonstrates the substantially better performance of the
;; analyzing evaluator.
